#include "HM_Reloc.h"

BOOL readPEInfo(char *modulePos, MZHeader *outMZ, PE_Header *outPE, PE_ExtHeader *outpeXH, SectionHeader **outSecHdr)
{
	MZHeader *mzH;
	mzH = (MZHeader *)modulePos;

	if(mzH->signature != 0x5a4d)
		return FALSE;

	PE_Header *peH;
	peH = (PE_Header *)(modulePos + mzH->offsetToPE);

	if(peH->sizeOfOptionHeader != sizeof(PE_ExtHeader))
		return FALSE;

	PE_ExtHeader *peXH;
	peXH = (PE_ExtHeader *)((char *)peH + sizeof(PE_Header));

	SectionHeader *secHdr = (SectionHeader *)((char *)peXH + sizeof(PE_ExtHeader));

	*outMZ = *mzH;
	*outPE = *peH;
	*outpeXH = *peXH;
	*outSecHdr = secHdr;

	return TRUE;
}

/*BOOL RelocImage(PVOID exeAddr, PVOID newAddr)
{
	MZHeader mzH2;
	PE_Header peH2;
	PE_ExtHeader peXH2;
	SectionHeader *secHdr2;

	if (!exeAddr || !newAddr)
		return FALSE;

	if(!readPEInfo((char *)exeAddr, &mzH2, &peH2, &peXH2, &secHdr2))
		return FALSE;

	if(peXH2.relocationTableAddress && peXH2.relocationTableSize) {
		FixupBlock *fixBlk = (FixupBlock *)((char *)exeAddr + peXH2.relocationTableAddress);		

		while(fixBlk->blockSize) {
			int numEntries = (fixBlk->blockSize - sizeof(FixupBlock)) >> 1;
			unsigned short *offsetPtr = (unsigned short *)(fixBlk + 1);
			for(int i = 0; i < numEntries; i++)	{				
				int relocType = (*offsetPtr & 0xF000) >> 12;
				if(relocType == 3) {
					DWORD *codeLoc = (DWORD *)((char *)exeAddr + fixBlk->pageRVA + (*offsetPtr & 0x0FFF));					
					DWORD delta = (DWORD)newAddr - (DWORD)peXH2.imageBase;					
					DWORD value = (*codeLoc) + delta;
					DWORD dummy;
					//if (!XXXX cambiare in memcpy HM_SafeWriteProcessMemory(GetCurrentProcess(), codeLoc, &value, sizeof(DWORD), &dummy))
					//	return FALSE;
				}
				offsetPtr++;
			}
			fixBlk = (FixupBlock *)offsetPtr;
		}
		return TRUE;
	}
	return FALSE;
}*/

// Per non usare la GetProcAddress che potrebbe avere una entry
// modificata nella import table
void* HM_SafeGetProcAddress(HMODULE hModule, char *func_to_search)
{
	BYTE *ImageBase = (BYTE *)hModule;
	WORD *PeOffs;
	IMAGE_NT_HEADERS *PE_Header;
	MY_IMAGE_EXPORT_DESCRIPTOR *Dll_Export;
	DWORD Index;
	unsigned short *Ordinal;
	DWORD *pFuncName;
	DWORD *pFunc_Pointer;

	// Verifica che sia un PE
	if (ImageBase[0]!='M' || ImageBase[1]!='Z')
		return NULL;

	PeOffs = (WORD *)&(ImageBase[0x3C]);
	PE_Header = (IMAGE_NT_HEADERS *)(ImageBase + (*PeOffs));
	// Qualche controllo sugli headers
	if (PE_Header->Signature != 0x00004550 || PE_Header->OptionalHeader.NumberOfRvaAndSizes < 1 ||
		PE_Header->OptionalHeader.DataDirectory[0].VirtualAddress == 0) 
		return NULL;

	Dll_Export = (MY_IMAGE_EXPORT_DESCRIPTOR *) (ImageBase + PE_Header->OptionalHeader.DataDirectory[0].VirtualAddress);
	// Scorre la lista di DLL importate
	for (Index=0; Index < Dll_Export->NumberOfNames ; Index++) {	
		pFuncName = (DWORD *)(ImageBase + Dll_Export->AddressOfNames + Index*4);
		if (*pFuncName == NULL)
			continue;			
		// Vede se e' la funzione che cerchiamo
		if (!strcmp(func_to_search, (char *)(ImageBase + *pFuncName)))
			break;
	}

	if(Index >= Dll_Export->NumberOfNames)
		return NULL;

	// Legge Ordinale
	Ordinal = (unsigned short *) (ImageBase + Dll_Export->AddressOfNameOrdinals + Index*2);
	// Legge il puntatore a funzione
	pFunc_Pointer = (DWORD *) (ImageBase + Dll_Export->AddressOfFunctions + (*Ordinal)*4);
	return (ImageBase + *pFunc_Pointer);
}

//*******************************************************************************************************
// Returns the total size required to load a PE image into memory
//
//*******************************************************************************************************

int calcTotalImageSize(MZHeader *inMZ, PE_Header *inPE, PE_ExtHeader *inpeXH,
					   SectionHeader *inSecHdr)
{
	int result = 0;
	int alignment = inpeXH->sectionAlignment;

	if(inpeXH->sizeOfHeaders % alignment == 0)
		result += inpeXH->sizeOfHeaders;
	else
	{
		int val = inpeXH->sizeOfHeaders / alignment;
		val++;
		result += (val * alignment);
	}
	for(int i = 0; i < inPE->numSections; i++)
	{
		if(inSecHdr[i].virtualSize)
		{
			if(inSecHdr[i].virtualSize % alignment == 0)
				result += inSecHdr[i].virtualSize;
			else
			{
				int val = inSecHdr[i].virtualSize / alignment;
				val++;
				result += (val * alignment);
			}
		}
	}

	return result;
}


//*******************************************************************************************************
// Returns the aligned size of a section
//
//*******************************************************************************************************

ULONG getAlignedSize(unsigned long curSize, unsigned long alignment)
{	
	if(curSize % alignment == 0)
		return curSize;
	else
	{
		int val = curSize / alignment;
		val++;
		return (val * alignment);
	}
}

//*******************************************************************************************************
// Copy a PE image from exePtr to ptrLoc with proper memory alignment of all sections
//
//*******************************************************************************************************

BOOL loadPE(char *exePtr, MZHeader *inMZ, PE_Header *inPE, PE_ExtHeader *inpeXH,
			SectionHeader *inSecHdr, LPVOID ptrLoc)
{
	char *outPtr = (char *)ptrLoc;

	memcpy(outPtr, exePtr, inpeXH->sizeOfHeaders);
	outPtr += getAlignedSize(inpeXH->sizeOfHeaders, inpeXH->sectionAlignment);

	for(int i = 0; i < inPE->numSections; i++)
	{
		if(inSecHdr[i].sizeOfRawData > 0)
		{
			unsigned long toRead = inSecHdr[i].sizeOfRawData;
			if(toRead > inSecHdr[i].virtualSize)
				toRead = inSecHdr[i].virtualSize;

			memcpy(outPtr, exePtr + inSecHdr[i].pointerToRawData, toRead);

			outPtr += getAlignedSize(inSecHdr[i].virtualSize, inpeXH->sectionAlignment);
		}
	}

	return true;
}


//*******************************************************************************************************
// Loads the DLL into memory and align it
//
//*******************************************************************************************************

LPVOID loadDLL(char *dllName)
{
	char moduleFilename[MAX_PATH + 1];
	LPVOID ptrLoc = NULL;
	MZHeader mzH2;
	PE_Header peH2;
	PE_ExtHeader peXH2;
	SectionHeader *secHdr2;

	GetSystemDirectory(moduleFilename, MAX_PATH);
	if((myStrlenA(moduleFilename) + myStrlenA(dllName)) >= MAX_PATH)
		return NULL;

	strncat_s(moduleFilename, MAX_PATH, dllName, MAX_PATH);

	// load this EXE into memory because we need its original Import Hint Table

	HANDLE fp;
	fp = CreateFile(moduleFilename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);

	if(fp != INVALID_HANDLE_VALUE)
	{
		BY_HANDLE_FILE_INFORMATION fileInfo;
		GetFileInformationByHandle(fp, &fileInfo);

		DWORD fileSize = fileInfo.nFileSizeLow;
		//printf("Size = %d\n", fileSize);
		if(fileSize)
		{
			//LPVOID exePtr = HeapAlloc(GetProcessHeap(), 0, fileSize);
			LPVOID exePtr = HM_SafeVirtualAllocEx(GetCurrentProcess(), NULL, fileSize, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
			if(exePtr)
			{
				DWORD read;

				if(ReadFile(fp, exePtr, fileSize, &read, NULL) && read == fileSize)
				{					
					if(readPEInfo((char *)exePtr, &mzH2, &peH2, &peXH2, &secHdr2))
					{
						int imageSize = calcTotalImageSize(&mzH2, &peH2, &peXH2, secHdr2);						

						ptrLoc = HM_SafeVirtualAllocEx(GetCurrentProcess(), NULL, imageSize, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
						//ptrLoc = HeapAlloc(GetProcessHeap(), 0, imageSize);
						if(ptrLoc)
						{							
							loadPE((char *)exePtr, &mzH2, &peH2, &peXH2, secHdr2, ptrLoc);
						}
					}

				}
				//HeapFree(GetProcessHeap(), 0, exePtr);
				VirtualFree(exePtr, 0, MEM_RELEASE);
			}
		}
		CloseHandle(fp);
	}

	return ptrLoc;
}

DWORD myStrlenA(char *ptr)
{
	DWORD len = 0;
	while(*ptr)
	{
		len++;
		ptr++;
	}

	return len;
}